/* 
 * FILE --- mlp.cc
 *
 * PURPOSE --- Source file for a class that implements a Multilayer 
 * Perceptron (MLP) neural network (NN). 
 * The user specifies the number of inputs, the number of outouts and the 
 * number of the neurons in the hidden layer of the perceptron. It is possible
 * to choose different activation functions for the neurans. The user can 
 * either specify the input and output weigth matrices and biaces, or let them 
 * be choosen randomly within the [-1.0, 1.0] range.
 *
 * This utility uses the matrix tool for matrix maltiplications.
 *
 */


// Include Files
#include <math.h>
#include "mlp.hh"
#include "random.hh"

// Define this macro if you want to use this utility without linking
// to the RCS Library
#ifdef NO_RCS
#include <stdio.h>            /* printf function */
#else
#include "rcs_prnt.hh"          /* rcs_print_ functions */
#endif


/*************************************************************/

MLP::MLP(int _num_of_inputs, int _num_of_outputs, int _num_of_neurons)
{
  // No error at start
  mlp_err = MLP_NO_ERROR;
  num_of_inputs = _num_of_inputs;
  num_of_outputs = _num_of_outputs;
  num_of_neurons = _num_of_neurons;

  // Allocate memory 
  allocate_memory();

  // If any error then report
  if (mlp_err != MLP_NO_ERROR)
    {
#ifdef NO_RCS
      printf("MLP: ERROR --- Not Enought Memory!!!\n");
      printf("The MLP network will not function correctly ...\n");
#else
      rcs_print("MLP: ERROR --- Not Enought Memory!!!\n");
      rcs_print("The MLP network will not function correctly ...\n");
#endif
    }

  // Initialize the gain and bias matrices with random numbers b/n [-1.0, 1.0]
  Randomize();
  set_input_weights(-1.0, 1.0);
  set_input_biases(-1.0, 1.0);
  set_output_weights(-1.0, 1.0);
  set_output_biases(-1.0, 1.0);

  // Set the activation function type as Logaritmic-Sigmoid
  set_activation_function_type(LOGSIG);
}
  
void MLP::allocate_memory(void)
{
  // Create the matrix of input weigths
  Wi = new matrix(num_of_neurons, num_of_inputs);
  if (Wi->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create the vector of input biases
  bi = new matrix(num_of_neurons, 1);
  if (bi->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create the matrix of output weigths
  Wo = new matrix(num_of_outputs, num_of_neurons);
  if (Wo->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create the vector of output biases
  bo = new matrix(num_of_outputs, 1);
  if (bo->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create input vector u
  u = new matrix(num_of_inputs, 1);
  if (u->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create intermediate vector x
  x = new matrix(num_of_neurons, 1);
  if (x->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create intermediate vector f
  phi = new matrix(num_of_neurons, 1);
  if (phi->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create output vector y
  y = new matrix(num_of_outputs, 1);
  if (y->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }

  // Create error vector e
  e = new matrix(num_of_outputs, 1);
  if (e->matrix_err != MATRIX_NO_ERROR)
    {
      mlp_err = MLP_MEM_ERROR;
    }
}


MLP::~MLP()
{
  // Free the allocated memory
  free_memory();
}


void MLP::free_memory(void)
{
  delete Wi;
  delete bi;
  delete Wo;
  delete bo;

  delete u;
  delete x;
  delete phi;
  delete y;
  delete e;
}


void MLP::set_input_weights(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the input gain matrix to random values b/n [_low, _high]
  for (ii=0; ii < num_of_neurons; ii++)
    {
      for (jj=0; jj < num_of_inputs; jj++)
	{
	  // Generate a random number b/n [_low, _high]
	  rand_val = Rand_Double(_low, _high);
	  
	  // Assign the element to the generated value
	  Wi->set_element(ii, jj, rand_val);
	}
    }
}


void MLP::set_input_weights(const double _in_weights)
{
  // Set the input gain matrix to the specified value
  Wi->set_matrix(_in_weights);
}


void MLP::set_input_weights(const double *_in_weights)
{
  // Set the input gain matrix to the specified value
  Wi->set_matrix(_in_weights);
}


void MLP::set_input_weights(const double **_in_weights)
{
  // Set the input gain matrix to the specified value
  Wi->set_matrix(_in_weights);
}


void MLP::set_input_biases(double _low, double _high)
{
  int ii;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the input bias vector to random values b/n [_low, _high]
  for (ii=0; ii < num_of_neurons; ii++)
    {
      // Generate a random number b/n [_low, _high]
      rand_val = Rand_Double(_low, _high);
      
      // Assign the element to the generated value
      bi->set_element(ii, 0, rand_val);
    }
}

void MLP::set_input_biases(const double _in_biases)
{
  // Set the input bias vector to the specified value
  bi->set_matrix(_in_biases);
}

void MLP::set_input_biases(const double *_in_biases)
{
  // Set the input bias vector to the specified value
  bi->set_matrix(_in_biases);
}



void MLP::set_output_weights(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the output gain matrix to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_neurons; jj++)
	{
	  // Generate a random number b/n [_low, _high]
	  rand_val = Rand_Double(_low, _high);
	  
	  // Assign the element to the generated value
	  Wo->set_element(ii, jj, rand_val);
	}
    }
}


void MLP::set_output_weights(const double _out_weights)
{
  // Set the output gain matrix to the specified value
  Wo->set_matrix(_out_weights);
}


void MLP::set_output_weights(const double *_out_weights)
{
  // Set the output gain matrix to the specified value
  Wo->set_matrix(_out_weights);
}


void MLP::set_output_weights(const double **_out_weights)
{
  // Set the output gain matrix to the specified value
  Wo->set_matrix(_out_weights);
}


void MLP::set_output_biases(double _low, double _high)
{
  int ii;  // Index variales for the loops
  double rand_val;  // Temporary variable

  // Set the output bias vector to random values b/n [_low, _high]
  for (ii=0; ii < num_of_outputs; ii++)
    {
      // Generate a random number b/n [_low, _high]
      rand_val = Rand_Double(_low, _high);
      
      // Assign the element to the generated value
      bo->set_element(ii, 0, rand_val);
    }
}


void MLP::set_output_biases(const double _out_biases)
{
  // Set the output bias vector to the specified value
  bo->set_matrix(_out_biases);
}


void MLP::set_output_biases(const double *_out_biases)
{
  // Set the output bias vector to the specified value
  bo->set_matrix(_out_biases);
}


void MLP::set_activation_function_type(int _actv_func_type)
{
  actv_func_type = _actv_func_type;

  switch(actv_func_type)
    {
    case LOGSIG:  // Log-Sigmoid activation function
      actv_func = logsig;
      break;

    case TANSIG:  // Hyperbolic Tangent Sigmoid activation function
      actv_func = tansig;
      break;

    case HARDLIM:  // Hard limiting function (like a step)
      actv_func = hardlim;
      break;

    case HARDLIMS:  // Hard limiting function (like a step) symmetric wrt origin
      actv_func = hardlims;
      break;

    case PURELIN:  // Purely linear function
      actv_func = purelin;
      break;

    case SATLIN:  // Linear function with saturation 
      actv_func = satlin;
      break;

    case SATLINS:  // Linear function with saturation symmetric wrt origin
      actv_func = satlins;
      break;

    case POSLIN:  // Positive linear function
      actv_func = poslin;
      break;

    default:
#ifdef NO_RCS
      printf("Invalid activation function type!!!\n");
      printf("Logaritmic-Sigmoidal function assumed ...\n");
#else
      rcs_print("Invalid activation function type!!!\n");
      rcs_print("Logaritmic-Sigmoidal function assumed ...\n");
#endif
      actv_func_type = LOGSIG;
      actv_func = logsig;
    }
}


void MLP::control(const double *input, double *_output)
{
  // Use matrix operations
  u->set_col(0, input);

  // Calculate the output of the input layer
  (*x) = (*Wi) * (*u) + (*bi);

  // Calculate the output of the activation functions
  get_activation();

  // Calculate the output of the output layer
  (*y) = (*Wo) * (*phi) + (*bo);
  
  // Return the output
  get_output(_output);
}


void MLP::update_output_params(const double *error)
{
  double temp;  // Temporory variable

  // Use matrix operations
  e->set_col(0, error);

  // This is used in both of the adaptations below
  temp = eta/(2.0 + phi_norm);

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm

  // Update the output weights
  (*Wo) = (*Wo) + temp * (*e) * transpose((*phi));

  // Update the output biases
  (*bo) = (*bo) + temp * (*e);
}


void MLP::update_output_params(const double *error, double norm2)
{
  double temp;  // Temporory variable

  // Use matrix operations
  e->set_col(0, error);

  // This is used in both of the adaptations below
  temp = eta/(3.0 + norm2 + phi_norm);

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm

  // Update the output weights
  (*Wo) = (*Wo) + temp * (*e) * transpose((*phi));

  // Update the output biases
  (*bo) = (*bo) + temp * (*e);
}


void MLP::update_output_params(const double *error, double norm2, double uu)
{
  double temp;  // Temporory variable

  // Use matrix operations
  e->set_col(0, error);

  // This is used in both of the adaptations below
  temp = (eta*uu)/(2.0 + norm2 + (uu*uu)*(1.0 + phi_norm));

  // Update the parameters based on the Normilized Gradient Method
  // or the so called Projection Algorithm

  // Update the output weights
  (*Wo) = (*Wo) + temp * (*e) * transpose((*phi));

  // Update the output biases
  (*bo) = (*bo) + temp * (*e);
}


void MLP::bound_output_weights(double _low, double _high)
{
  int ii, jj;  // Index variales for the loops

  // Set the output gain matrix to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      for (jj = 0; jj < num_of_neurons; jj++)
	{
	  if (Wo->get_element(ii, jj) < _low)
	    {
	      // Assign the element to the lower bound
	      Wo->set_element(ii, jj, _low);
	    }
	  else if (Wo->get_element(ii, jj) > _high)
	    {
	      // Assign the element to the upper bound
	      Wo->set_element(ii, jj, _high);
	    }
	}
    }
}


void MLP::bound_output_biases(double _low, double _high)
{
  int ii;  // Index variales for the loops

  // Set the output bias vector to random values b/n [_low, _high]
  for (ii = 0; ii < num_of_outputs; ii++)
    {
      if (bo->get_element(ii, 0) < _low)
	{
	  // Assign the element to the lower bound
	  bo->set_element(ii, 0, _low);
	}
      else if (bo->get_element(ii, 0) > _high)
	{
	  // Assign the element to the upper bound
	  bo->set_element(ii, 0, _high);
	}
    }
}


void MLP::get_activation(void)
{
  double func_val;  // Temporary variable

  // Calculate the activation of each neuron
  for(int ii = 0; ii < num_of_neurons; ii++)
    {
      func_val = actv_func(x->get_element(ii, 0));
      phi->set_element(ii, 0, func_val);
    }

  // Calculate the norm of phi (needed in indirect adaptive control)
  phi_norm = (transpose((*phi)) * (*phi)).get_element(0, 0);
}


void MLP::get_output(double *_output)
{
  // Return the output
  for(int ii = 0; ii < num_of_outputs; ii++)
    {
      _output[ii] = y->get_element(ii, 0);
    }
}


// Activation Functions

inline double logsig(const double _input)
{
  return ( (1.0 / (1.0 + exp(-_input))) );
}


inline double tansig(const double _input)
{
  return ( ((exp(_input) - exp(-_input)) / (exp(_input) + exp(-_input))) );
}

inline double hardlim(const double _input)
{
  return ( _input < 0.0 ? 0.0 : 1.0 );
}


inline double hardlims(const double _input)
{
  return ( _input < 0.0 ? -1.0 : 1.0 );
}

inline double purelin(const double _input)
{
  return ( _input );
}

inline double satlin(const double _input)
{
  return ( _input < 0.0 ? 0.0 : (_input > 1.0 ? 1.0 : _input) );
}

inline double satlins(const double _input)
{
  return ( _input < -1.0 ? -1.0 : (_input > 1.0 ? 1.0 : _input) );
}

inline double poslin(const double _input)
{
  return (_input < 0.0 ? 0.0 : _input );
}

